/**\file SimulationDecPOMDPDiscrete.cpp
 *
 * Authors:
 * Frans Oliehoek <faolieho@science.uva.nl>
 * Matthijs Spaan <mtjspaan@isr.ist.utl.pt>
 *
 * Copyright 2008 Universiteit van Amsterdam, Instituto Superior Tecnico
 *
 * This file is part of MultiAgentDecisionProcess.
 *
 * MultiAgentDecisionProcess is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * MultiAgentDecisionProcess is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiAgentDecisionProcess.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * $Id: SimulationDecPOMDPDiscrete.cpp 2151 2008-01-21 16:32:52Z mtjspaan $
 */

#include "SimulationDecPOMDPDiscrete.h"
#include "JointPolicyDiscrete.h"
#include "AgentLocalObservations.h"
#include "AgentSharedObservations.h"
#include "AgentFullyObservable.h"

#include "JointObservationHistoryTree.h"
#include "JointObservation.h"
#include "JointAction.h"

using namespace std;

SimulationDecPOMDPDiscrete::
SimulationDecPOMDPDiscrete(const PlanningUnitDecPOMDPDiscrete &pu, 
                           int nrRuns, int seed) : 
    Simulation(nrRuns, seed),
    _m_pu(&pu),
    _m_saveIntermediateResults(false)
{
}

SimulationDecPOMDPDiscrete::
SimulationDecPOMDPDiscrete(const PlanningUnitDecPOMDPDiscrete &pu, 
                           const ArgumentHandlers::Arguments &args) : 
    Simulation(args.nrRuns, args.randomSeed),
    _m_pu(&pu),
    _m_saveIntermediateResults(false)
{
    if(args.verbose)
        SetVerbose(true);
}

//Destructor
SimulationDecPOMDPDiscrete::~SimulationDecPOMDPDiscrete()
{
}

SimulationResult SimulationDecPOMDPDiscrete::Initialize() const
{
    SimulationResult result(_m_pu->GetHorizon(),GetRandomSeed(),GetNrRuns());

    if(GetVerbose())
        cout << "Simulation::RunSimulations horizon " << _m_pu->GetHorizon()
             << " nrRuns " << GetNrRuns() << " seed " 
             << GetRandomSeed() << endl;
    
    if(GetRandomSeed()!=illegalRandomSeed)
    {
        // Seed the random number generator
        srand(GetRandomSeed());
    }

    return(result);
}

void SimulationDecPOMDPDiscrete::SaveIntermediateResults(string filename)
{
    _m_saveIntermediateResults=true;
    _m_intermediateResultsFilename=filename;
}

SimulationResult
SimulationDecPOMDPDiscrete::RunSimulations(const JointPolicyDiscrete *jp) const
{
    SimulationResult result=Initialize();

    if(GetVerbose())
        jp->Print();

    // Run the simulations
    int i;
    for(i=0;i<GetNrRuns();i++)
        result.AddReward(RunSimulation(jp));

    return(result);
}

double
SimulationDecPOMDPDiscrete::RunSimulation(const JointPolicyDiscrete *jp) const
{
    Index jaI,sI,joI;
    unsigned int h;
    double r,sumR=0;
    JointObservationHistoryTree* joht = 
        _m_pu->GetJointObservationHistoryTree(
            INITIAL_JOHI);
    
    sI = _m_pu->GetReferred()->SampleInitialState();

    if(GetVerbose())
        cout << "Simulation::RunSimulation set initial state to " 
             << sI << endl;

    for(h=0;h<_m_pu->GetHorizon();h++)
    {	
        Index johI = joht->GetIndex();
        //JointPolicyDiscrete can be stochastic...
        //jaI = jp->GetJointActionIndex(johI);
        jaI = jp->SampleJointAction(johI);

        Step(jaI, h, sI, joI, r, sumR);
        
        /* action taken at ts=0,...,hor-1 - therefore only observation
         * histories at ts=0,...,hor-2 have successors.*/
        if(h < _m_pu->GetHorizon()-1) 
            joht = joht->GetSuccessor(joI);
    }

    return(sumR);
}

void SimulationDecPOMDPDiscrete::Step(Index jaI, unsigned int t, Index &sI,
                                      Index &joI, double &r,
                                      double &sumR) const
{
    Index sI_suc=_m_pu->GetReferred()->SampleSuccessorState(sI,jaI);

    joI=_m_pu->GetReferred()->SampleJointObservation(jaI,sI_suc);
    r = _m_pu->GetReferred()->GetReward(sI,jaI);
        
    // calc. the discounted reward
    sumR+=r*pow(_m_pu->GetDiscount(),static_cast<double>(t));
    
    if(GetVerbose())
        cout << "Simulation::RunSimulation ("
             << sI << "," << jaI << "," << sI_suc << ") ("
             << (_m_pu->GetReferred()->GetState(sI)->SoftPrintBrief())
             << ","
             << (_m_pu->GetJointAction(jaI))->SoftPrintBrief()
             << ","
             << (_m_pu->GetReferred()->GetState(sI_suc)->SoftPrintBrief())
             << ") (p " 
             << _m_pu->GetReferred()->GetTransitionProbability(sI,jaI,sI_suc) 
             << ") jo " << joI << " "
             << (_m_pu->GetJointObservation(joI))->SoftPrintBrief()
             << " (p " 
             << _m_pu->GetReferred()->GetObservationProbability(jaI,sI_suc,joI)
             << ") r " << r << " sumR " << sumR << endl;

    sI = sI_suc;
}

Index SimulationDecPOMDPDiscrete::
GetAction(const vector<AgentLocalObservations*> &agents, Index i,
          Index jaI, Index joI, double r, Index prevJoI, Index sI) const
{
    vector<Index> oIs=_m_pu->JointToIndividualObservationIndices(joI);
    return(agents[i]->Act(oIs[i]));
}

Index SimulationDecPOMDPDiscrete::
GetAction(const vector<AgentSharedObservations*> &agents, Index i,
          Index jaI, Index joI, double r, Index prevJoI, Index sI) const
{
    return(agents[i]->Act(joI));
}

Index SimulationDecPOMDPDiscrete::
GetAction(const vector<AgentFullyObservable*> &agents,
          Index i, Index jaI, Index joI, double r, Index prevJoI,
          Index sI) const
{
    return(agents[i]->Act(sI,joI,r));
}
