/**\file ValueFunctionDecPOMDPDiscrete.cpp
 *
 * Authors:
 * Frans Oliehoek <faolieho@science.uva.nl>
 * Matthijs Spaan <mtjspaan@isr.ist.utl.pt>
 *
 * Copyright 2008 Universiteit van Amsterdam, Instituto Superior Tecnico
 *
 * This file is part of MultiAgentDecisionProcess.
 *
 * MultiAgentDecisionProcess is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * MultiAgentDecisionProcess is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiAgentDecisionProcess.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * $Id: ValueFunctionDecPOMDPDiscrete.cpp 2116 2008-01-16 23:11:49Z mtjspaan $
 */

#include "ValueFunctionDecPOMDPDiscrete.h"

#define DEBUG_CALCV 0
#define DEBUG_CALCV_CACHE 0

using namespace std;

ValueFunctionDecPOMDPDiscrete::ValueFunctionDecPOMDPDiscrete(
    PlanningUnitDecPOMDPDiscrete& p , JointPolicyDiscretePure& jp):
    Referrer<PlanningUnitDecPOMDPDiscrete>(&p),
    Referrer<JointPolicyDiscretePure>(&jp)
{
    _m_V_initialized = false;
    _m_p_V = 0;
}

ValueFunctionDecPOMDPDiscrete::ValueFunctionDecPOMDPDiscrete(
    PlanningUnitDecPOMDPDiscrete* p , JointPolicyDiscretePure* jp):
    Referrer<PlanningUnitDecPOMDPDiscrete>(p),
    Referrer<JointPolicyDiscretePure>(jp)
{
    _m_V_initialized = false;
    _m_p_V = 0;
}

//Copy assignment constructor.    
ValueFunctionDecPOMDPDiscrete::ValueFunctionDecPOMDPDiscrete(
    const ValueFunctionDecPOMDPDiscrete& o) 
{
    _m_V_initialized = false;
    _m_p_V = new Matrix(*o._m_p_V);//this makes a deep copy.
}

//Destructor
ValueFunctionDecPOMDPDiscrete::~ValueFunctionDecPOMDPDiscrete()
{
    DeleteV();
}

void ValueFunctionDecPOMDPDiscrete::DeleteV()
{
    if(_m_V_initialized)
    {
        delete(_m_p_V);
        _m_V_initialized = false;
    }
    _m_cached.clear();
}

void ValueFunctionDecPOMDPDiscrete::CreateV()
{
    if(_m_V_initialized)
        DeleteV();

    _m_p_V = new Matrix(GetPU()->GetNrStates(),
                        GetPU()->GetNrJointObservationHistories());
    _m_p_V->clear();
    _m_V_initialized = true;
}

bool ValueFunctionDecPOMDPDiscrete::IsCached(Index sI, Index JOHI)
{
    return(_m_cached[GetCombinedIndex(sI, JOHI)]);
}

void ValueFunctionDecPOMDPDiscrete::SetCached(Index sI, Index JOHI)
{
    pair<Index, bool> p(GetCombinedIndex(sI, JOHI), true);
    pair<map<Index,bool>::iterator, bool> result = _m_cached.insert( p );
    if(result.second != true)
    {
        // can't insert if this key ( GetCombinedIndex(sI, JOHI) ) is already
        // present - in this case, we need to change the value to true...
        // *(result.first) isa pair<int,bool>, so we do:
        (*(result.first)).second = true;
    }
}

double ValueFunctionDecPOMDPDiscrete::CalculateV() 
{
    return CalculateV0Recursively(true);
}

double ValueFunctionDecPOMDPDiscrete::CalculateV0Recursively(bool cache)
{
    if(cache)
    {
        CreateV();
        if(DEBUG_CALCV_CACHE)
            cout << "CalculateV0Recursively::creating a new value function"
                 << endl;
    }
    if(DEBUG_CALCV) {cout << "evaluating joint policy:\n"; GetJPol()->Print();}

    Index nrS = GetPU()->GetNrStates(); 
    double val = 0;

    for(Index sI = 0; sI < nrS; sI++)
    {
        double v_sI = CalculateVsjohRecursively(sI, Globals::INITIAL_JOHI,
                                                cache);
        if(DEBUG_CALCV){
            cout << ">>>ValueFunctionDecPOMDPDiscrete::CalculateV() -"
                 << " CalculateVsjohRecursively(sI=" << sI
                 << ", INITIAL_JOHI, cache="<< cache<< ") = " << v_sI << endl; }
        val +=  GetPU()->GetInitialStateProbability(sI) * v_sI;
    }
    if(DEBUG_CALCV) {  cout << "This policy's V=" << val <<endl;}
    return val;
}

double ValueFunctionDecPOMDPDiscrete::CalculateVsjohRecursively(Index sI, 
    Index johI, bool cache)
{
    if(DEBUG_CALCV)
    {cout<< "\nValueFunctionDecPOMDPDiscrete::CalculateVsjohRecursively("
         << sI << ", " << johI << ") called"<<endl;}

    if(cache && IsCached(sI, johI))
    {
        if(DEBUG_CALCV_CACHE)cout << "returning cached result"<<endl;
        return(    (*_m_p_V)(sI, johI) );
    }
    Index jaI = GetJPol()->GetJointActionIndex(johI);

    double R = GetPU()->GetReward(sI, jaI);
    double ExpFutureR = 0.0;
    size_t h = GetPU()->GetHorizon();
    // horizon h policy makes decision on observation histories which 
    // maximally have length h-1. 
    // t=0 - () - length=0 
    // ... 
    // t=h-1 - (o1,...,o{h-1}) - length h-1
    if(GetPU()->GetTimeStepForJOHI(johI) >= h - 1 )
    {
        if(DEBUG_CALCV){ 
            cout << "ValueFunctionDecPOMDPDiscrete::CalculateVsjoh"
                 << "Recursively("<< sI <<", " << johI << ") - V="<<R<<endl;} 
        return(R);
    }

    if(DEBUG_CALCV){ cout << "Calculating future reward"<<endl;}
    Index nrS = GetPU()->GetNrStates();
    for(Index sucSI = 0; sucSI < nrS; sucSI++)
    {
        double probSucSI = GetPU()->GetTransitionProbability(sI, jaI,
               sucSI);
        if(DEBUG_CALCV){ cout << "P(s"<<sucSI<<"|s"<<sI<<",ja"<<jaI<<")="<<
                probSucSI<<endl;}
        Index nrJO = GetPU()->GetNrJointObservations();
        for(Index joI = 0; joI < nrJO; joI++)
        {
            double probJOI =  GetPU()->GetObservationProbability(jaI,
               sucSI, joI);
            if(DEBUG_CALCV){ cout << "P(jo"<<joI<<"|ja"<<jaI<<",s"<<sucSI<<")="
                                  <<probJOI<<endl;}

            Index sucJohI = GetPU()->GetSuccessorJOHI(johI, joI);
        
            double thisSucV = CalculateVsjohRecursively(sucSI, sucJohI, cache);

            ExpFutureR += probSucSI * probJOI * thisSucV;
        }//end for each observation
    }//end for each potential succesor state
    double val = R + ExpFutureR;
    if(cache)
    {
        (*_m_p_V)(sI, johI) = val;
        SetCached(sI, johI);
    }
    if(DEBUG_CALCV){ 
        cout << "ValueFunctionDecPOMDPDiscrete::CalculateVsjohRecursively("
             << sI <<", " << johI << ") \n->immediate R="<<R<<
            " \n->exp. future reward="<<ExpFutureR<<"\n->V="<<val<<endl;} 
    return(val);
}
