/**\file JointPolicyPureVector.cpp
 *
 * Authors:
 * Frans Oliehoek <faolieho@science.uva.nl>
 * Matthijs Spaan <mtjspaan@isr.ist.utl.pt>
 *
 * Copyright 2008 Universiteit van Amsterdam, Instituto Superior Tecnico
 *
 * This file is part of MultiAgentDecisionProcess.
 *
 * MultiAgentDecisionProcess is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * MultiAgentDecisionProcess is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiAgentDecisionProcess.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * $Id: JointPolicyPureVector.cpp 2151 2008-01-21 16:32:52Z mtjspaan $
 */

#include "JointPolicyPureVector.h"
#include "IndexTools.h"

using namespace std;

#define DEBUG_JPPT 0 
#define DEBUG_JPPT_GETJA_LOCAL 0

JointPolicyPureVector::JointPolicyPureVector(const 
        Interface_ProblemToPolicyDiscretePure& pu)
    : JointPolicyDiscretePure( pu , pu.GetDefaultIndexDomCat()  )
{
    size_t nrA = pu.GetNrAgents();
    for(Index agentI = 0; agentI < nrA; agentI++)
    {
        PolicyPureVector* pol = new PolicyPureVector(
            *GetInterfacePTPDiscretePure(), agentI, pu.GetDefaultIndexDomCat());
        _m_policyPointerVector.push_back(pol);
    }
}

JointPolicyPureVector::JointPolicyPureVector(
        const Interface_ProblemToPolicyDiscretePure& pu,
        PolicyGlobals::IndexDomainCategory idc )
    : JointPolicyDiscretePure( pu , idc )
{
    if(DEBUG_JPPT) cout << "JointPolicyPureVector::JointPolicyPureVector"<<
        "(Interface_ProblemToPolicyDiscretePure& pu) - new joint policy" << endl;

    for(Index agentI = 0; agentI < GetNrAgents(); agentI++)
    {
        PolicyPureVector* pol = new PolicyPureVector(
                *GetInterfacePTPDiscretePure(), agentI, idc);
        _m_policyPointerVector.push_back(pol);
    }
}


//Copy constructor.    
JointPolicyPureVector::JointPolicyPureVector(const JointPolicyPureVector& o)
    :
    JointPolicyDiscretePure( o )
{
    if(DEBUG_JPPT) cout << " clone JointPolicyPureVector constructor ";
    
    //copy constructor for superclass should be called so this is unecess.
    //SetReferred(o.GetInterfacePTPDiscretePure()); //just copies pointer.
    
    // we want to create an actual clone of this joint policy and thus
    // of the member (individual policies)
    vector<PolicyPureVector*>::const_iterator it = 
    o._m_policyPointerVector.begin();
    vector<PolicyPureVector*>::const_iterator last = 
    o._m_policyPointerVector.end();

    while(it != last)
    {
        //make a copy of the individual policy
        PolicyPureVector* ind_pol = new PolicyPureVector(**it);
        //store the reference to the new copy 
        _m_policyPointerVector.push_back(ind_pol);
        it++;
    }
       
}

//Destructor
JointPolicyPureVector::~JointPolicyPureVector()
{
    //prevent memory leakage... should delete the individual policies referred 
    //to...
    ClearIndividualPolicies();
}

void JointPolicyPureVector::ClearIndividualPolicies()
{
    //NOTE: this means that when adding a ind. policy p to the joint policy (or 
    //replacing one) we should make a copy of p and insert the ref to this copy.
    //Otherwise deleting the joint policy deletes the individual policies, 
    //whose other references become dangling.
    vector<PolicyPureVector*>::iterator it = _m_policyPointerVector.begin();
    vector<PolicyPureVector*>::iterator last = _m_policyPointerVector.end();

    while(it != last)
    {
        delete (*it);
        it++;
    }
    _m_policyPointerVector.clear();    

}

//TODO: check this code
JointPolicyPureVector& JointPolicyPureVector::operator= (JointPolicyPureVector& jp)
{
if(DEBUG_JPPT)     cout << "JointPolicyPureVector::operator= copy assignment operator called"<<endl;
    if (this == &jp) return *this;//handle self assingment

    //do assignment at parent level:
    JointPolicyDiscretePure::operator=(jp);
    //old:
/*    //just copies pointer.
//    Referrer<Interface_ProblemToPolicyDiscretePure>::
        SetReferred(
            jp.//Referrer<Interface_ProblemToPolicyDiscretePure>::
            GetInterfacePTPDiscretePure() ); 
*/

    vector<PolicyPureVector*>::iterator it = _m_policyPointerVector.begin();
    vector<PolicyPureVector*>::iterator last =  _m_policyPointerVector.end();
    vector<PolicyPureVector*>::iterator it2 = jp._m_policyPointerVector.begin();
    vector<PolicyPureVector*>::iterator last2 = jp._m_policyPointerVector.end();
    
    //TODO check if we really want to delete the PolicyPureVectors pointed to..?
    //I guess yes - but then this means that we should not be able to add indiv.
    //policies without copying them!!! make sure of this
    if(_m_policyPointerVector.size() != jp._m_policyPointerVector.size() )
    {
if(DEBUG_JPPT){ cout << "JointPolicyPureVector::operator=  - _m_policyPointerVector elements do not have same sizes!!!" << endl; cout << "this = "; this->Print(); cout << "jp = "; jp.Print();    }
    while(it != last)
    { 
        delete(*it);
        it++;
    }
    _m_policyPointerVector.clear();
    while(it2 != last2)
    {
        PolicyPureVector* p = new PolicyPureVector(**it2);//copy constructor
        _m_policyPointerVector.push_back(p);
        it2++;
    }
    }
    else //TODO test this construction...
    {
    while(it != last)
    { 
        **it = **it2; // PolicyPureVector = PolicyPureVector -> calls copy 
            // assignment operator of PolicyPureVector
        it++;it2++;
    }
    
    }
if(DEBUG_JPPT){ cout << "\nJointPolicyPureVector::operator= finished - result:";
    this->Print();}
    
    return(*this);
    
}

void JointPolicyPureVector::SetInterface_PTPDiscretePure(
        Interface_ProblemToPolicyDiscretePure& pu)
{
    SetReferred(&pu);
}
void JointPolicyPureVector::RandomInitialization()
{
    if(GetInterfacePTPDiscretePure() == 0)
    {
        cout << "JointPolicyPureVector:RandomInitialization:Error!!! GetInterfacePTPDiscretePure() not initialized yet" << endl;
    }
    for(Index agentI = 0; agentI < GetInterfacePTPDiscretePure()->GetNrAgents(); agentI++)
        _m_policyPointerVector[agentI]->RandomInitialization();
}
void JointPolicyPureVector::RandomInitialization(Index agentI)
{
    if(GetInterfacePTPDiscretePure() == 0)
    {
        cout << "JointPolicyPureVector:RandomInitialization:Error!!! GetInterfacePTPDiscretePure() not initialized yet" << endl;
    }
    if(agentI < GetInterfacePTPDiscretePure()->GetNrAgents())
        _m_policyPointerVector[agentI]->RandomInitialization();
}

bool JointPolicyPureVector::ZeroInitialization()
{
    if(GetInterfacePTPDiscretePure() == 0)
    {
        cout << "JointPolicyPureVector:ZeroInitialization:Error!!! GetInterfacePTPDiscretePure() not initialized yet" << endl;
        return(false);
    }
    for(Index agentI = 0; agentI < GetInterfacePTPDiscretePure()->GetNrAgents(); agentI++)
        _m_policyPointerVector[agentI]->ZeroInitialization();
    return(true);
}
bool JointPolicyPureVector::ZeroInitialization(Index agentI)
{
    if(GetInterfacePTPDiscretePure() == 0)
    {
        cout << "JointPolicyPureVector:ZeroInitialization:Error!!! GetInterfacePTPDiscretePure() not initialized yet" << endl;
        return(false);
    }
    if(agentI < GetInterfacePTPDiscretePure()->GetNrAgents())
        _m_policyPointerVector[agentI]->ZeroInitialization();
    return(true);
}

bool JointPolicyPureVector::Increment()
{
    bool carry_over = true;
    size_t nrAgents = GetNrAgents();
    Index agentI = nrAgents - 1;
    while(carry_over /*&& agentI >= 0*/)
    {
//    _m_obsHistToActionIndices[i] = (_m_obsHistToActionIndices[i] + 1) % nrA;
    carry_over = ++(*_m_policyPointerVector[agentI]) ;
        //(_m_obsHistToActionIndices[i] == 0);
    if(agentI >=1)
            agentI--;
        else//agentI is 0; -- will get huge unsigned int leading to seg.fault
            //if we don't jump out
            break;
    }
    return(carry_over);
}

LIndex JointPolicyPureVector::GetIndex() const
{
    int nrAgents=GetNrAgents();

    vector<LIndex> nrElems(2);
    nrElems[0]=1;
    nrElems[1]=1;

    vector<LIndex> indexVec(2);

    vector<LIndex> nrIndPol(nrAgents);
    vector<LIndex> individualIndices(nrAgents);
    for(int k=0;k<nrAgents;++k)
    {
        nrIndPol[k]=GetInterfacePTPDiscretePure()->GetNrPolicies(k,
                GetIndexDomainCategory() );
        individualIndices[k]=_m_policyPointerVector[k]->GetIndex();
    }

    LIndex i=individualIndices[0];

    for(int k=1;k<nrAgents;++k)
    {
        
        nrElems[0]*=nrIndPol[k-1];
        nrElems[1]=nrIndPol[k-1];

        indexVec[0]=i;
        indexVec[1]=individualIndices[k];
        i=IndexTools::IndividualToJointIndices(indexVec,nrElems);
#if 0
        cout << SoftPrintVector(nrIndPol) << " "
             << SoftPrintVector(individualIndices)
             << " " << k << " "
             << SoftPrintVector(indexVec) << " "
             << SoftPrintVector(nrElems) << endl;
#endif
    }
    return(i);
}

void JointPolicyPureVector::SetIndex(LIndex i)
{
    int nrAgents=GetNrAgents();

    vector<LIndex> nrElems(2);
    nrElems[0]=1;
    nrElems[1]=1;

    vector<LIndex> indexVec(2);

    vector<LIndex> nrIndPol(nrAgents);
    for(int k=0;k<nrAgents;++k)
    {
        nrIndPol[k]=GetInterfacePTPDiscretePure()->GetNrPolicies(k,
                GetIndexDomainCategory() );
        nrElems[0]*=nrIndPol[k];
    }

    for(int k=nrAgents-1;k>0;--k)
    {
        nrElems[0]*=nrIndPol[k-1];
        nrElems[1]=nrIndPol[k-1];

        indexVec=IndexTools::JointToIndividualIndices(i,nrElems);
        _m_policyPointerVector[k]->SetIndex(indexVec[1]);
        i=indexVec[0];
#if 0
        cout << SoftPrintVector(nrIndPol)
             << " " << k << " "
             << SoftPrintVector(indexVec) << " "
             << SoftPrintVector(nrElems) << endl;
#endif
    }
    _m_policyPointerVector[0]->SetIndex(indexVec[0]);
}

/* this is an old version - because this is called often(!), we replaced vectors
 * by arrays.
Index JointPolicyPureVector::GetJointActionIndex(Index johI) const
{
    if(DEBUG_JPPT_GETJA_LOCAL) {cout << "JointPolicyPureVector::GetJointActio"<<
        "nIndex(Index johI="<<johI<<") called"<<endl;}

    size_t nrA = GetInterfacePTPDiscretePure()->GetNrAgents();
    vector<Index> individualOHindices = 
        GetInterfacePTPDiscretePure()->JointToIndividualObservationHistoryIndices(johI);

    if( individualOHindices.size() != 
        static_cast<vector<Index>::size_type>(nrA)   )
    {
        individualOHindices.clear();
        stringstream ss;
        ss<< " JointPolicyPureVector::GetJointActionIndex Dimension msimatch"<<
            "johI has " << individualOHindices.size() <<" individual actions "<<
            "but there are "<< nrA <<" agents!\n";
        throw E(ss);
    }

    vector<Index> individualActionIndices;
    for( Index agentI=0; agentI < nrA; agentI++ )
    {
        if(DEBUG_JPPT_GETJA_LOCAL) {cout << "JointPolicyPureVector::GetJoint"<<
            "ActionIndex - Getting action for agent "<< agentI <<
            " observation history " << individualOHindices[agentI] << "..." << 
            endl;}

        Index aThisAgent = _m_policyPointerVector[agentI]->GetActionIndex(
            individualOHindices[agentI]);
        individualActionIndices.push_back(aThisAgent);
    }
    if(DEBUG_JPPT_GETJA_LOCAL){cout << "JointPolicyPureVector::GetJointAction"<<
        "Index - individualActionIndices.size()=" << individualActionIndices.
        size() << endl;}

    Index ja_i = GetInterfacePTPDiscretePure()->IndividualToJointActionIndices(
            individualActionIndices);

    return(ja_i);
}
*/
Index JointPolicyPureVector::GetJointActionIndex(Index jdI) const
{
#if DEBUG_JPPT_GETJA_LOCAL
    cout << "JointPolicyPureVector::GetJointActio"<<
        "nIndex(Index jdI="<<jdI<<") called"<<endl;
#endif
    size_t nrA = 0;
   
#if 1 
    // it is possible to cache the joint->individual observ. history mapping
    // If this is done we want to use it by calling 
    //    JointToIndividualObservationHistoryIndicesRef
    // as it is faster.
    // Otherwise we have to use 
    //    JointToIndividualObservationHistoryIndices
    // to avoid a segfault.
    //
    // The mentioned cache is calculated, when generating joint observation
    // histories. This is controlled by the 
    // PlanningUnitMADPDiscreteParameters of the PlanningUnitMADPDiscrete.
    // ( _m_params.GetComputeJointObservationHistories() in particular )
    vector<Index> individualDomainIndices;

    //if the "Interface_ProblemToPolicyDiscretePure"-implementing object this
    //policy caches joint -> indiv OH indices
    if(GetInterfacePTPDiscretePure()->AreCachedJointToIndivIndices(
                this->GetIndexDomainCategory() ) ) 
    {
        individualDomainIndices = GetInterfacePTPDiscretePure()->
                JointToIndividualPolicyDomainIndicesRef(jdI, 
                        this->GetIndexDomainCategory());
    } 
    else 
    { 
        individualDomainIndices = GetInterfacePTPDiscretePure()->
                JointToIndividualPolicyDomainIndices(jdI,
                        this->GetIndexDomainCategory());
    }
    nrA = GetInterfacePTPDiscretePure()->GetNrAgents(); 
    if( individualDomainIndices.size() !=
            static_cast<vector<Index>::size_type>(nrA)   ) 
    {
        stringstream ss; 
        ss<<" JointPolicyPureVector::GetJointActionIndex Dimension msimatch"<< 
            "jdI has " << individualDomainIndices.size() <<"individual actions "<< 
            "but there are "<< nrA <<" agents!\n"; 
        throw E(ss); 
    }
#endif

    nrA = individualDomainIndices.size();
        //GetInterfacePTPDiscretePure()->GetNrAgents();

    vector<Index> individualActionIndices(nrA);
    for( Index agentI=0; agentI < nrA; ++agentI )
    {
        Index ohI=individualDomainIndices[agentI];
        Index aThisAgent = _m_policyPointerVector[agentI]->GetActionIndex(ohI);
        individualActionIndices[agentI]=aThisAgent;
    }
    Index ja_i = GetInterfacePTPDiscretePure()->IndividualToJointActionIndices(
            individualActionIndices);

    return(ja_i);
}

void JointPolicyPureVector::SetDepth(size_t d)
{
    JointPolicy::SetDepth(d);

    vector<PolicyPureVector*>::iterator it = _m_policyPointerVector.begin();
    vector<PolicyPureVector*>::iterator last = _m_policyPointerVector.end();

    while(it != last)
    {
        (*it)->SetDepth(d);
        it++;
    }
}

string JointPolicyPureVector::SoftPrint(void) const
{
    if(GetInterfacePTPDiscretePure() == 0 || _m_policyPointerVector.size() != 
        static_cast< vector<PolicyPureVector*>::size_type> (GetInterfacePTPDiscretePure()->GetNrAgents()) )
    {
        throw E("JointPolicyPureVector:SoftPrint - Error!!! GetInterfacePTPDiscretePure() or _m_policyPointerVector not initialized yet");
    }
    stringstream ss;
    ss << "JointPolicyPureVector index " << GetIndex() << " depth " 
       << GetDepth() << endl;
    for(
            Index agentI = 0; 
            agentI <    static_cast< vector<PolicyPureVector*>::size_type> 
                            (GetInterfacePTPDiscretePure()->GetNrAgents());
            agentI++
        )
    {
        ss << "Policy for agent " << agentI << " (index "
           << _m_policyPointerVector[agentI]->GetIndex() << "):" << endl
           << _m_policyPointerVector[agentI]->SoftPrint();
    }
    return(ss.str());
} 

string JointPolicyPureVector::SoftPrintBrief(void) const
{
    if(GetInterfacePTPDiscretePure() == 0 || _m_policyPointerVector.size() != 
        static_cast< vector<PolicyPureVector*>::size_type> (GetInterfacePTPDiscretePure()->GetNrAgents()) )
    {
        throw E("JointPolicyPureVector:SoftPrintBrief - Error!!! GetInterfacePTPDiscretePure() or _m_policyPointerVector not initialized yet");
    }

    stringstream ss;
    ss << GetIndex();

    return(ss.str());
}
