/**\file JointPolicyPureVector.h
 *
 * Authors:
 * Frans Oliehoek <faolieho@science.uva.nl>
 * Matthijs Spaan <mtjspaan@isr.ist.utl.pt>
 *
 * Copyright 2008 Universiteit van Amsterdam, Instituto Superior Tecnico
 *
 * This file is part of MultiAgentDecisionProcess.
 *
 * MultiAgentDecisionProcess is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * MultiAgentDecisionProcess is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiAgentDecisionProcess.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * $Id: JointPolicyPureVector.h 2151 2008-01-21 16:32:52Z mtjspaan $
 */

/* Only include this header file once. */
#ifndef _JOINTPOLICYPUREVECTOR_H_
#define _JOINTPOLICYPUREVECTOR_H_ 1

/* the include directives */
#include <iostream>
#include <vector>
#include "Globals.h"
#include "PolicyPureVector.h"
#include "JointActionDiscrete.h"
#include "JointPolicyDiscretePure.h"


/// JointPolicyPureVector represents a discrete pure joint policy.
/** 
 * Each of the individual policies is a PolicyPureVector - a
 * deterministic mapping from observations to actions implemented by a
 * tree. This class acts as a wrapper. It contains a vector with
 * pointers to indiv. policies.  */
class JointPolicyPureVector  :  public JointPolicyDiscretePure
{
    private:

    /** \brief Private function that deletes all the individual
     * policies pointed to.*/
    void ClearIndividualPolicies();
  
    protected:
        std::vector<PolicyPureVector*> _m_policyPointerVector;

    public:
        // Constructor, destructor and copy assignment.

        /**\brief Constructor. 
         *
         * Associates a problem (typically a planning unit) with the joint
         * policy.  Information regarding the problem is used to construct a
         * joint policy of the proper shape.
         * 
         * Use the default IndexDomainCategory defined by pu.
         * */
        JointPolicyPureVector(const Interface_ProblemToPolicyDiscretePure& pu);
        /**\brief Constructor. 
         *
         * Associates a problem (typically a planning unit) with the joint
         * policy.  Information regarding the problem is used to construct a
         * joint policy of the proper shape.
         *
         * idc is the IndexDomainCategory over which the policy is specified.
         *
         * */
        JointPolicyPureVector(const Interface_ProblemToPolicyDiscretePure& pu,
                PolicyGlobals::IndexDomainCategory idc );
        /// Copy constructor.
        JointPolicyPureVector(const JointPolicyPureVector& a);
        /// Destructor.
        ~JointPolicyPureVector();
        /**The copy assignment operator - makes deep copies using 
         * PolicyPureVector::operator= which makes deep copies.*/
        JointPolicyPureVector& operator= (JointPolicyPureVector& jp);

        //operators:
        ///Performs increment. See Increment().
        bool operator++ ()   {    return( this->Increment() );     };
        
        /// Get the index of this joint policy.
        LIndex GetIndex() const;

        /// Sets the index of this joint policy. Updates the policies represented.
        void SetIndex(LIndex i);

        //data manipulation (set) functions:
        ///Randomly initialize the joint policy.
        void RandomInitialization();
        ///Randomly initialize the policy for agentI.
        void RandomInitialization(Index agentI);
        /// Initialize the joint policy to the first joint policy.
        /** This is the joint policy that specifies action 0 --the
         *  first action-- for all observation histories.*/
        bool ZeroInitialization();
        /// Initialize the policy of agentI to the first policy.
        /** This is the policy that specifies action 0 --the first
         *  action-- for all observation histories.*/
        bool ZeroInitialization(Index agentI);
        ///Sets the planning unit for this joint policy
        inline void SetInterface_PTPDiscretePure(
                Interface_ProblemToPolicyDiscretePure& pu);
        ///Increments the individual policy of agentI.
        bool Increment(Index agentI) 
        {return( ++(*_m_policyPointerVector[agentI])) ;};
        ///Increments the joint policy, used to iterate over joint policies.
        bool Increment();

        /// Sets the depth of the joint policy.
        /** Also sets the depth of each of the invididual policies. */
        void SetDepth(size_t d);

        ///Sets the policy for agentI to domainI->aI.
        void SetAction(Index agentI, Index domainI, Index aI)
            {_m_policyPointerVector[agentI]->SetAction(domainI, aI);}

        //get (data) functions:
        
        //other functions
        
        /// Prints a description of this JointPolicyPureVector to a string.
        std::string SoftPrint() const; 

        /// Prints a brief description to a string.
        std::string SoftPrintBrief() const; 
        void PrintBrief() const
        { std::cout << SoftPrintBrief();}
        
        
        ///Returns the jaI taken by this policy for joint domain index johI.
        Index GetJointActionIndex(Index i) const;
        ///Returns the action index for domainI for agent aI.
        Index GetActionIndex(Index aI, Index domainI) const
            {return _m_policyPointerVector[aI]->GetActionIndex(domainI);}
      
};


#endif /* !_JOINTPOLICYPUREVECTOR_H_ */


// Local Variables: ***
// mode:c++ ***
;// End: ***
