/**\file PolicyDiscrete.h
 *
 * Authors:
 * Frans Oliehoek <faolieho@science.uva.nl>
 * Matthijs Spaan <mtjspaan@isr.ist.utl.pt>
 *
 * Copyright 2008 Universiteit van Amsterdam, Instituto Superior Tecnico
 *
 * This file is part of MultiAgentDecisionProcess.
 *
 * MultiAgentDecisionProcess is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * MultiAgentDecisionProcess is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiAgentDecisionProcess.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * $Id: PolicyDiscrete.h 2116 2008-01-16 23:11:49Z mtjspaan $
 */

/* Only include this header file once. */
#ifndef _POLICYDISCRETE_H_
#define _POLICYDISCRETE_H_ 1

/* the include directives */
#include <iostream>
#include "Globals.h"
#include "Policy.h"
#include "Interface_ProblemToPolicyDiscrete.h"
#include "Referrer.h"

using namespace PolicyGlobals;

/**\brief PolicyDiscrete is a class that represents a discrete 
 * policy.
 *
 * A `discrete joint policy' is a policy for a discrete problem. I.e., the 
 * problem specifies, for each agent, a discrete domain for the policy and 
 * discrete actions.
 *
 * A PolicyDiscrete from discrete indices over the domain (typically 
 * action-observation histories or types) to (degenerate) probability 
 * distributions over indices over (joint) actions.
 *
 * This means that this class includes both pure and stochastic pure policies.
 *
 * */
class PolicyDiscrete : public Referrer<Interface_ProblemToPolicyDiscrete>,
                            public Policy
{
    private:    
        /// Maintains the IndexDomainCategory.
        PolicyGlobals::IndexDomainCategory _m_indexDomCat;
    
    protected:
    
    public:
        // Constructor, destructor and copy assignment.
        /// (default) Constructor
        PolicyDiscrete(const Interface_ProblemToPolicyDiscrete& iptpd,
                       PolicyGlobals::IndexDomainCategory idc ) :
            Referrer<Interface_ProblemToPolicyDiscrete>::Referrer(
                const_cast<Interface_ProblemToPolicyDiscrete&>(iptpd)),
            _m_indexDomCat(idc){};
            
        /// Copy constructor.
        PolicyDiscrete(const PolicyDiscrete& a) :
            Referrer<Interface_ProblemToPolicyDiscrete>::Referrer(a.GetReferred()),
            _m_indexDomCat(a._m_indexDomCat){};

        /// Destructor.
        virtual ~PolicyDiscrete()
        {};

        //operators:

        //data manipulation (set) functions:
        /**\brief Sets the category of the domain over which the indices of
         * this policy are specified.
         */
        void SetIndexDomainCategory(IndexDomainCategory idc)
        {_m_indexDomCat = idc;};
        
        //get (data) functions:
        /**\brief Returns the Category of the domain over which the indices of
         * this policy are specified.
         */
        IndexDomainCategory GetIndexDomainCategory() const
        {return (_m_indexDomCat);};

        /**\brief Returns the probability that the policy specifies joint action
         * for joint domain index i.
         *
         * Implementations for pure policies clearly should return 0 or 1.
         */
        virtual double GetActionProb( Index i, Index ja ) const = 0;
        
        /**\brief Returns the probability that the policy specifies joint action
         * for joint domain index i.
         *
         * Implementations for pure policies clearly should return 0 or 1.
         */
        double SampleAction( Index i, Index ja ) const;

        /**\brief return a pointer to the referred 
         * Interface_ProblemToPolicyDiscrete.
         */
        const Interface_ProblemToPolicyDiscrete* GetInterfacePTPDiscrete() const
        {return GetReferred();}

};


#endif /* !_POLICYDISCRETE_H_ */


// Local Variables: ***
// mode:c++ ***
// End: ***
