/**\file argumentHandlers.cpp
 *
 * Authors:
 * Frans Oliehoek <faolieho@science.uva.nl>
 * Matthijs Spaan <mtjspaan@isr.ist.utl.pt>
 *
 * Copyright 2008 Universiteit van Amsterdam, Instituto Superior Tecnico
 *
 * This file is part of MultiAgentDecisionProcess.
 *
 * MultiAgentDecisionProcess is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * MultiAgentDecisionProcess is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiAgentDecisionProcess.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * $Id: argumentHandlers.cpp 2151 2008-01-21 16:32:52Z mtjspaan $
 */

#include "argumentHandlers.h"

using namespace ArgumentHandlers;

namespace ArgumentHandlers {

const char *argp_program_bug_address = 
"<faolieho@science.uva.nl>, <mtjspaan@isr.ist.utl.pt>";

// a group for input arguments
static const int GID_INPUTARG=2;


//Dec-POMDP File options (problemFile)
static const int GID_PROBLEMFILE=GID_INPUTARG;
const char *problemFile_argp_version = ".dpomdp file argument parser 0.1";
static const char *problemFile_args_doc = "PROBLEM";
static const char *problemFile_doc = 
"This is the documentation for the .dpomdp file argument parser\
This parser should be included as a child argp parser in the \
main argp parser of your application.(and this message will\
not be shown)\
\vPROBLEM is a Dec-POMDP problem filename (.dpomdp).";

static struct argp_option problemFile_options[] = {
{ 0 }
};
error_t
problemFile_parse_argument (int key, char *arg, struct argp_state *state)
{
    static bool got_problemFile = 0;
    //if we already got the dec-pomdp file return
    if(got_problemFile)
        return ARGP_ERR_UNKNOWN;
    /* Get the input argument from argp_parse, which we
      know is a pointer to our arguments structure. */
    struct Arguments *theArgumentsStruc = (struct Arguments*) state->input;

    switch (key)
    {
        case ARGP_KEY_NO_ARGS:
            argp_usage (state);
            break;
        case ARGP_KEY_ARG:
            theArgumentsStruc->dpf = arg;
            got_problemFile = 1;
            break;
        default:
            return ARGP_ERR_UNKNOWN;
     }
    return 0;
}
static struct argp problemFile_argp = { problemFile_options, problemFile_parse_argument, problemFile_args_doc, problemFile_doc };
extern const struct argp_child problemFile_child = {&problemFile_argp, 0, 0, GID_PROBLEMFILE};



//global options (globalOptions)
static const int GID_GLOBALOPTIONS=8;
const char *globalOptions_argp_version = "global options parser 0.1";
static const char *globalOptions_args_doc = 0;
static const char *globalOptions_doc = 
"This is the documentation for the global options parser\
This parser should be included as a child argp parser in the \
main argp parser of your application. (and this message will\
not be shown)"; 
//\v";
static struct argp_option globalOptions_options[] = {
{"verbose",  'v', 0,       0, "Produce verbose output" },
{"quiet",    'q', 0,       0, "Don't produce any output" },
{"silent",   's', 0,       OPTION_ALIAS },
{ 0 }
};
error_t
globalOptions_parse_argument (int key, char *arg, struct argp_state *state)
{
    /* Get the input argument from argp_parse, which we
      know is a pointer to our arguments structure. */
    struct Arguments *theArgumentsStruc = (struct Arguments*) state->input;

    switch (key)
    {
        case 'q': case 's':
            theArgumentsStruc->quiet = 1;
            break;
        case 'v':
            theArgumentsStruc->verbose = 1;
            break;
        default:
            return ARGP_ERR_UNKNOWN;
     }
    return 0;
}
static struct argp globalOptions_argp = { globalOptions_options, globalOptions_parse_argument, globalOptions_args_doc, globalOptions_doc };
const struct argp_child globalOptions_child = {&globalOptions_argp, 0, "General options", GID_GLOBALOPTIONS};



//model options modelOptions
static const int GID_MODELOPTIONS=3;
const char *modelOptions_argp_version = "Model options parser 0.1";
static const char *modelOptions_args_doc = 0;
static const char *modelOptions_doc = 
"This is the documentation for the model options parser\
This parser should be included as a child argp parser in the \
main argp parser of your application. (and this message will\
not be shown)\
\v";

static const int OPT_TOI=1;
static struct argp_option modelOptions_options[] = {
{"sparse",  's', 0,       0, "Use sparse transition and observation models" },
{ 0 }
};
error_t
modelOptions_parse_argument (int key, char *arg, struct argp_state *state)
{
    /* Get the input argument from argp_parse, which we
      know is a pointer to our arguments structure. */
    struct Arguments *theArgumentsStruc = (struct Arguments*) state->input;

    switch (key)
    {
        case 's': 
            theArgumentsStruc->sparse = 1;
            break;
        default:
            return ARGP_ERR_UNKNOWN;
     }
    return 0;
}
static struct argp modelOptions_argp = { modelOptions_options, modelOptions_parse_argument, modelOptions_args_doc, modelOptions_doc };
const struct argp_child modelOptions_child = {&modelOptions_argp, 0, "Model options", GID_MODELOPTIONS};


//Solution method options (solutionMethodOptions)
const char *solutionMethodOptions_argp_version = "Solution method options parser 0.1";
static const char *solutionMethodOptions_args_doc = 0;
static const char *solutionMethodOptions_doc = 
"This is the documentation for the solution method options parser\
This parser should be included as a child argp parser in the \
main argp parser of your application. (and this message will\
not be shown)"; 
//\v";

//we define a special solution method group ID, so it is easier to group other
//parsers with this one
static const int GID_SM=2;
static struct argp_option solutionMethodOptions_options[] = {
{"horizon",'h',"HOR", 0, "Specifies the planning horizon (defaults to 2)" },
{ 0 }
};
error_t
solutionMethodOptions_parse_argument (int key, char *arg, struct argp_state *state)
{
    /* Get the input argument from argp_parse, which we
      know is a pointer to our arguments structure. */
    struct Arguments *theArgumentsStruc = (struct Arguments*) state->input;

    switch (key)
    {
        case 'h': 
            theArgumentsStruc->horizon = atoi(arg);
            break;
        default:
            return ARGP_ERR_UNKNOWN;
     }
    return 0;
}
static struct argp solutionMethodOptions_argp = { solutionMethodOptions_options, solutionMethodOptions_parse_argument, solutionMethodOptions_args_doc, solutionMethodOptions_doc };
const struct argp_child solutionMethodOptions_child = {&solutionMethodOptions_argp, 0, "Solution method options", GID_SM };


//joint policy argument parser (jpolIndex)
static const int GID_JPOLINDEX=GID_INPUTARG;
const char *jpolIndex_argp_version = "joint policy index argument parser 0.1";
static const char *jpolIndex_args_doc = "JPOL-INDEX";
static const char *jpolIndex_doc = 
"This is the documentation for the joint policy index argument parser\
This parser should be included as a child argp parser in the \
main argp parser of your application.(and this message will\
not be shown)\
\vJPOL-INDEX is a (long long) integer that is the index of the joint \
policy to be considered.";
static struct argp_option jpolIndex_options[] = {
//{"none",  'v', 0,       0, "no options" },
{ 0 }
};
error_t
jpolIndex_parse_argument (int key, char *arg, struct argp_state *state)
{
    static bool got_jpolIndex = 0;
    //if we already got the dec-pomdp file return
    if(got_jpolIndex)
        return ARGP_ERR_UNKNOWN;
    /* Get the input argument from argp_parse, which we
      know is a pointer to our arguments structure. */
    struct Arguments *theArgumentsStruc = (struct Arguments*) state->input;

    switch (key)
    {
        case ARGP_KEY_NO_ARGS:
            argp_usage (state);
            break;
        case ARGP_KEY_ARG:
        {
            char *left;
            theArgumentsStruc->jpolIndex = strtoull(arg,&left,10);
            if(*left != '\0')
            {
                std::stringstream ss;
                ss << "Could not entirely parse the JPOL-INDEX: arg='"
                   << arg <<"', left='" << left
                   << "', found jpol index='"<< theArgumentsStruc->jpolIndex
                   << "'";
                throw(E(ss));
            }
            got_jpolIndex = 1;
            break;
        }
        default:
            return ARGP_ERR_UNKNOWN;
     }
    return 0;
}
static struct argp jpolIndex_argp = { jpolIndex_options, jpolIndex_parse_argument, jpolIndex_args_doc, jpolIndex_doc };
const struct argp_child jpolIndex_child = {&jpolIndex_argp, 0, 0, GID_PROBLEMFILE};

//Simulation options (simulation)
static const int GID_SIMULATION=7;
const char *simulation_argp_version = "Simulation options parser 0.1";
static const char *simulation_args_doc = 0;
static const char *simulation_doc = 
"This is the documentation for the options options parser\
This parser should be included as a child argp parser in the \
main argp parser of your application. (and this message will\
not be shown)"; 
//\v";
static struct argp_option simulation_options[] = {
{"runs",  'r', "RUNS", 0, "Set the number of episodes to simulate" },
{"seed",  'S', "SEED", 0, "Set the random seed" },
{ 0 }
};
error_t
simulation_parse_argument (int key, char *arg, struct argp_state *state)
{
    /* Get the input argument from argp_parse, which we
      know is a pointer to our arguments structure. */
    struct Arguments *theArgumentsStruc = (struct Arguments*) state->input;

    switch (key)
    {
    case 'r':
        theArgumentsStruc->nrRuns=atoi(arg);
        break;
    case 'S':
        theArgumentsStruc->randomSeed=atoi(arg);
        break;
    default:
        return ARGP_ERR_UNKNOWN;
    }
    return 0;
}
static struct argp simulation_argp = { simulation_options,
                                       simulation_parse_argument,
                                       simulation_args_doc, simulation_doc };
const struct argp_child simulation_child = {&simulation_argp, 0,
                                            "Simulation options", 
                                            GID_SIMULATION};

} // end namespace
